package org.bitbucket.bradleysmithllc.etlunit.feature.database;

import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.util.DimensionalHashMap;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.util.DimensionalMap;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.io.file.*;
import org.bitbucket.bradleysmithllc.json.validator.ResourceNotFoundException;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.*;
import java.sql.Date;
import java.util.*;

public class DatabaseClassListener extends NullClassListener implements ExtractOperationProcessor, StageOperationProcessor, TruncateOperationProcessor, MigrateOperationProcessor
{
	public static final String DEFAULT_MODE = null;
	public static final String LAST_EXTRACT_FILE_SCHEMA = "lastExtractFileSchema";
	private final DatabaseFeatureModule parent;
	private final DatabaseConfiguration databaseConfiguration;
	private DatabaseRuntimeSupport databaseRuntimeSupport;
	private RuntimeSupport runtimeSupport;
	protected Log applicationLog;
	private JDBCClient jdbcClient;

	private boolean prepareDatabase = true;
	private boolean refreshStageData = false;
	private boolean forceInitializeDatabase = false;

	private DataFileManager dataFileManager;

	public DatabaseClassListener(DatabaseFeatureModule parent, DatabaseConfiguration databaseConfiguration)
	{
		this.parent = parent;
		this.databaseConfiguration = databaseConfiguration;
	}

	@Inject
	public void receiveDataFileManager(DataFileManager manager)
	{
		dataFileManager = manager;
	}

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport manager)
	{
		runtimeSupport = manager;
	}

	@Inject
	public void setExportStageData(@Named("database.refreshStageData") RuntimeOption option)
	{
		refreshStageData = option.isEnabled();

		applicationLog.info(refreshStageData ? "Refreshing stage data" : "Not refreshing stage data");
	}

	@Inject
	public void setPrepareDatabase(@Named("database.prepareDatabase") RuntimeOption option)
	{
		prepareDatabase = option.isEnabled();

		applicationLog.info(prepareDatabase ? "Preparing database" : "Not preparing database");
	}

	@Inject
	public void setForceInitializeDatabase(@Named("database.forceInitializeDatabase") RuntimeOption option)
	{
		forceInitializeDatabase = option.isEnabled();

		applicationLog.info(forceInitializeDatabase ? "Forcing database" : "Not forcing database");
	}

	@Inject
	public void setJdbcClient(JDBCClient client)
	{
		jdbcClient = client;
	}

	@Inject
	public void setApplicationLog(@Named("applicationLog") Log log)
	{
		applicationLog = log;
	}

	@Inject
	public void setDatabaseRuntimeSupport(DatabaseRuntimeSupport databaseRuntimeSupport)
	{
		this.databaseRuntimeSupport = databaseRuntimeSupport;
	}

	@Override
	public action_code process(ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		String name = etlTestOperation.getOperationName();

		if (name.equals("stage"))
		{
			throw new RuntimeException();
		}
		else if (name.equals("extract"))
		{
			throw new RuntimeException();
		}

		return action_code.defer;
	}

	@Override
	public void begin(final ETLTestMethod mt, VariableContext context) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		// if this method, or class is annotated with @NoDatabase, then no further work is done
		if (mt.getAnnotations("@NoDatabase").size() > 0 || mt.getTestClass().getAnnotations("@NoDatabase").size() > 0)
		{
			// do nothing
			return;
		}

		// create a view of the annotations which has them in priority order descending
		List<ETLTestAnnotation> annoList = new ArrayList<ETLTestAnnotation>(mt.getTestClass().getAnnotations("@Database"));
		annoList.addAll(mt.getAnnotations("@Database"));

		// before starting a method, initialize any database support necessary.
		// get a list of databases to prepare
		DimensionalMap<DatabaseConnection, String, Object> dbMap = new DimensionalHashMap<DatabaseConnection, String, Object>();

		if (databaseConfiguration == null)
		{
			throw new TestExecutionError("Database feature lacks configuration", DatabaseConstants.ERR_MISSING_DATABASE_CONFIGURATION);
		}

		Iterator<ETLTestAnnotation> it = annoList.iterator();

		while (it.hasNext())
		{
			ETLTestAnnotation anno = it.next();

			ETLTestValueObject annoet = anno.getValue();
			String id = annoet.query("id").getValueAsString();

			ETLTestValueObject modes = annoet.query("modes");

			DatabaseConnection dc = databaseConfiguration.getDatabaseConnection(id);

			if (dc == null)
			{
				throw new TestExecutionError("Undeclared connection id " + id);
			}

			// clear previous modes for this connection
			dbMap.remove(dc);

			if (modes != null)
			{
				List<String> list = modes.getValueAsListOfStrings();

				for (String mode : list)
				{
					dbMap.put(dc, mode, "");
				}
			}
			else
			{
				dbMap.put(dc, DEFAULT_MODE, "");
			}
		}

		if (dbMap.size() == 0)
		{
			// nothing to do at all
			return;
		}

		// prepare the databases
		Iterator<DatabaseConnection> set = dbMap.keySet().iterator();

		while (set.hasNext())
		{
			DatabaseConnection dc = set.next();

			// look up the implementation and request a prepare
			DatabaseImplementation impl = parent.getImplementation(dc.getId());

			if (impl == null)
			{
				throw new TestExecutionError("Connection id " + dc.getId() + " does not have a valid implementation: " + dc.getImplementationId(), DatabaseConstants.ERR_MISSING_IMPLEMENTATION);
			}

			Map<String, Object> lmap = dbMap.get(dc);

			for (final String mode : lmap.keySet())
			{
				// validate the modes
				// check that the configuration allows for any mode (indicated by no mode specified on the connection)
				final DatabaseConnection databaseConnection = databaseConfiguration.getDatabaseConnection(dc.getId());

				// potentially initialize
				if (prepareDatabase)
				{
					if (!parent.initializedDbs.containsKey(dc.getId(), mode))
					{
						applicationLog.info("Initializing database: " + dc.getId() + (mode == null ? "" : ("." + mode)));

						if (forceInitializeDatabase || parent.databaseOutOfDate(mode, databaseConnection) || impl.getDatabaseState(databaseConnection, mode) == DatabaseImplementation.database_state.fail)
						{
							DenyAllOperationsOperationRequest request = new DenyAllOperationsOperationRequest()
							{
								@Override
								public DatabaseImplementation.InitializeRequest getInitializeRequest() throws UnsupportedOperationException
								{
									return new DatabaseImplementation.InitializeRequest()
									{
										public String getMode()
										{
											return mode;
										}

										public DatabaseConnection getConnection()
										{
											return databaseConnection;
										}

										public ETLTestMethod getTestMethod()
										{
											return mt;
										}

										public File resolveSchemaReference(String ddlRef)
										{
											throw new UnsupportedOperationException();
										}
									};
								}
							};

							impl.processOperation(DatabaseImplementation.operation.createDatabase, request);

							jdbcClient.useStatement(databaseConnection, mode, new JDBCClient.StatementClient()
							{
								public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception
								{
									for (String ref : databaseConnection.getSchemaScripts())
									{
										applicationLog.info("Processing schema source: " + ref);

										SQLAggregator SQLAggregator = parent.resolveDDLRef(ref, databaseConnection);

										int count = 0;

										SQLAggregator.Aggregator agg = SQLAggregator.getStatementAggregator();

										while (agg.hasNext())
										{
											count++;

											SQLAggregator.FileRef fRef = agg.next();

											String sql = fRef.getLine();

											try
											{
												st.execute(sql);
											}
											catch (Throwable thr)
											{
												throw new TestExecutionError("Error processing DDL reference: " + fRef.describe(), DatabaseConstants.ERR_DDL_APPLICATION_FAILURE, thr);
											}
										}

										applicationLog.info("Processed [" + count + "] statements");
									}
								}
							});

							// let the implementation take the schema created and prepare for testing
							impl.processOperation(DatabaseImplementation.operation.completeDatabaseInitialization, request);
							parent.updateDatabaseFlag(mode, databaseConnection);
						}

						parent.initializedDbs.put(dc.getId(), mode, "");
					}

					// prepare the database for test
					applicationLog.info("Preparing database for test: " + dc.getId() + (mode == null ? "" : ("." + mode)));

					impl.processOperation(DatabaseImplementation.operation.prepareDatabase, new DenyAllOperationsOperationRequest()
					{
						@Override
						public DatabaseImplementation.PrepareRequest getPrepareRequest() throws UnsupportedOperationException
						{
							return new DatabaseImplementation.PrepareRequest()
							{
								public ETLTestMethod getTestMethod()
								{
									return mt;
								}

								public String getMode()
								{
									return mode;
								}

								public DatabaseConnection getConnection()
								{
									return databaseConnection;
								}
							};
						}
					});
				}
				else if (parent.databaseOutOfDate(mode, databaseConnection))
				{
					throw new TestExecutionError("Prepare database option is false, but the database schema has changed.", DatabaseConstants.ERR_INVALID_SCHEMA_STATE);
				}
			}
		}

		/* store database information in the variable context for use by other features
		 * These should look like this to the clients:
		 * database: {
		 * 	connections: {
		 * 		connectionName: {
		 * 			target: databaseConnection
		 * 		}
		 *	 }
		 *	}
		 *	Where the database connection is a facade which only responds to the configured
		 *	mode (target in the example).  The 'modeless' connection uses the
		 *	reserved name 'default'
		 */

		// create the database container
		ETLTestValueObjectBuilder builder = new ETLTestValueObjectBuilder();
		builder = builder.object().key("connections").object();

		// iterate through all required connections and modes
		for (DatabaseConnection key : dbMap.keySet())
		{
			Map<String, Object> lmap = dbMap.get(key);

			// create the base connection object
			// this will contain modes
			builder = builder.key(key.getId()).object();
			// run through all modes, checking for the special null mode which gets converted
			// to 'default'

			for (String mode : lmap.keySet())
			{
				if (mode == null)
				{
					builder.key("default").object();
				}
				else
				{
					builder.key(mode).object();
				}

				// add properties
				parent.propogateProperties(builder, key, mode);
				builder.endObject();
			}

			builder.endObject();
		}

		builder.endObject().endObject();

		context.declareAndSetValue("database", builder.toObject());
	}

	public action_code processExtract(ExtractOperation operation, ETLTestOperation op, final ETLTestValueObject obj, final VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		final String source;
		String apackage = op.getTestClass().getPackage();
		DatabaseImplementation.ExtractRequest.source_mode smode = DatabaseImplementation.ExtractRequest.source_mode.table;

		String sqlScript = operation.getSqlScript();
		String extractOperationSql = operation.getSql();

		if (sqlScript != null)
		{
			// load the sql into the source variable
			try
			{
				File sqlFile = databaseRuntimeSupport.getSourceScript(apackage, sqlScript);

				if (!sqlFile.exists())
				{
					throw new TestExecutionError("Sql script not found: " + sqlFile);
				}

				source = IOUtils.readFileToString(sqlFile);
				smode = DatabaseImplementation.ExtractRequest.source_mode.sql;
			}
			catch (IOException e)
			{
				throw new TestExecutionError("", e);
			}
		}
		else if (extractOperationSql != null)
		{
			smode = DatabaseImplementation.ExtractRequest.source_mode.sql;
			source = extractOperationSql;
		}
		else
		{
			source = operation.getSourceTable();
		}

		final String sourceTable = source;
		final String target = operation.getTarget();
		final DatabaseImplementation.ExtractRequest.source_mode sourcemode = smode;

		ExtractOperation.columnListMode_enum lm = operation.getColumnListMode();

		final List<String> list = operation.getColumnList();

		if (list != null && lm == null)
		{
			// exclude is the default
			lm = ExtractOperation.columnListMode_enum.exclude;
		}
		else if (lm != null && list == null)
		{
			throw new TestExecutionError("Extract operation does not allow " + ExtractOperation.COLUMNLISTMODE_JSON_NAME + " without " + ExtractOperation.COLUMNLIST_JSON_NAME);
		}

		final ExtractOperation.columnListMode_enum listMode = lm;

		if ((listMode != null || list != null) && sourcemode == DatabaseImplementation.ExtractRequest.source_mode.sql)
		{
			throw new TestExecutionError("Extract operation does not allow " + ExtractOperation.COLUMNLISTMODE_JSON_NAME + " using sql");
		}

		// look for optional data
		final String targetFile = operation.getTargetFile();

		ConnectionMode connMode = getConnection(parent, obj, context, op.getTestMethod()).get(0);

		final DatabaseConnection conn = connMode.conn;

		final String mode = connMode.modes.get(0);

		// now send it off to the implementation
		DatabaseImplementation impl = getDatabaseImplementation(parent, conn);

		String aschema = operation.getSourceSchema();

		if (aschema == null)
		{
			aschema = impl.getDefaultSchema(conn, mode);
		}

		final String schema = aschema;

		final File targetBase;

		if (targetFile != null)
		{
			File f = new File(targetFile);

			if (f.isAbsolute())
			{
				targetBase = f;
			}
			else
			{
				targetBase = databaseRuntimeSupport.getGeneratedDataSet(conn, connMode.getPrettyMode(), targetFile, DatabaseImplementation.data_format.delimited);
			}
		}
		else
		{
			targetBase = null;
		}

		final String key;

		if (sourcemode == DatabaseImplementation.ExtractRequest.source_mode.sql)
		{
			key = (conn.getId() + ".SQL." + source).toUpperCase();
		}
		else
		{
			key = (conn.getId() + "." + schema.toUpperCase() + "." + source).toUpperCase();
		}

		// get the relational table definition
		try
		{
			Map<String, RelationalDataSet> meta = getTableMetaData(conn, mode, impl, context);
			RelationalDataSet tr = meta.get(key);

			// load column and key data if not loaded already
			if (sourcemode == DatabaseImplementation.ExtractRequest.source_mode.table)
			{
				if (tr == null)
				{
					throw new TestExecutionError("Database does not export table " + key + " in it's meta data", DatabaseConstants.ERR_MISSING_SCHEMA_TABLE);
				}
			}
			else
			{
				// create this relational record
				tr = new RelationalDataSet(source, conn, impl, sqlScript);
				meta.put(key, tr);
			}

			final RelationalDataSet rt = tr;

			loadTable(rt, mode);

			final File dataTarget;

			if (targetFile == null)
			{
				dataTarget = databaseRuntimeSupport.getSourceDataSet(apackage, target, DatabaseImplementation.data_format.delimited);
			}
			else
			{
				dataTarget = targetBase;
			}

			applicationLog.info("Extracting source " + (sourcemode == DatabaseImplementation.ExtractRequest.source_mode.table ? "table" : "query") + "[" + key + "]" + " to local dataset [" + dataTarget.getAbsolutePath() + "]");

			// export the data
			jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient()
			{
				public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception
				{
					RelationalDataSet relationalDataSet = rt;

					DataFileSchema fileSchema = loadSchema(rt, source, obj);

					applicationLog.info("Resolved data file schema: " + fileSchema.getId());

					fileSchema.setColumnDelimiter(databaseConfiguration.getDataFileColumnDelimiter());
					fileSchema.setRowDelimiter(databaseConfiguration.getDataFileRowDelimiter());
					fileSchema.setNullToken(databaseConfiguration.getDataFileNullToken());

					if (listMode != null)
					{
						relationalDataSet = relationalDataSet.createSubView(list, listMode);
						switch (listMode)
						{
							case include:
								fileSchema = fileSchema.createSubViewIncludingColumns(list, "subIncluding");
								break;
							case exclude:
								fileSchema = fileSchema.createSubViewExcludingColumns(list, "subExcluding");
								break;
						}
					}

					// store the generated file schema for reference
					File fileGen = runtimeSupport.getGeneratedSourceDirectory("file");

					IOUtils.writeStringToFile(new FileBuilder(fileGen).subdir("synthetic-fml").mkdirs().name(fileSchema.getId()).file(), fileSchema.toJsonString());

					context.declareAndSetValue(LAST_EXTRACT_FILE_SCHEMA, new ETLTestValueObjectImpl(fileSchema));

					DataFile ff = dataFileManager.loadDataFile(dataTarget, fileSchema);

					String variableName = "extract.columns." + target;
					applicationLog.info("Setting context variable for " + variableName);
					context.declareAndSetValue(variableName, new ETLTestValueObjectImpl(relationalDataSet.getColumnNames()));

					String sqlSelect = relationalDataSet.createSQLSelect();
					applicationLog.info("Using sql select for object(" + key + " in [" + connection.getId() + "." + mode + "]):\n" + sqlSelect);

					long startTime = System.currentTimeMillis();
					int rowsRead = 0;

					DataFileWriter ffwriter = ff.getWriter(new JdbcTypeConverter());

					try
					{
						ResultSet rs = st.executeQuery(sqlSelect);

						ResultSetMetaData rsmd = rs.getMetaData();

						int cols = rsmd.getColumnCount();

						try
						{
							Map<String, Object> rowData = new HashMap<String, Object>();

							while (rs.next())
							{
								rowsRead++;

								for (int i = 1; i <= cols; i++)
								{
									String colName = rsmd.getColumnName(i);
									Object data = rs.getObject(i);

									rowData.put(colName, data);
								}

								ffwriter.addRow(rowData);
							}
						}
						finally
						{
							rs.close();
						}
					}
					finally
					{
						ffwriter.close();
					}

					long endTime = System.currentTimeMillis();
					applicationLog.info("Selected [" + rowsRead + "] rows in [" + (endTime - startTime) + "] ms.");
				}
			});
		}
		catch (TestExecutionError e)
		{
			throw e;
		}
		catch (Exception e)
		{
			throw new TestExecutionError("", e);
		}

		return action_code.handled;
	}

	private DataFileSchema loadSchema(RelationalDataSet rt, String source, ETLTestValueObject obj)
	{
		// look for the meta info in the project
		try
		{
			// look it up in the project
			return databaseRuntimeSupport.getRelationalMetaInfo(rt.getDatabaseConnection(), rt.getSchema(), rt.getName(), source, obj);
		}
		catch(ResourceNotFoundException exc)
		{
			// use the relational definition
			return rt.getFileSchema();
		}
	}

	private void loadTable(final RelationalDataSet rt, String mode) throws Exception
	{
		applicationLog.info("Updating database meta data for target [" + rt.getDatabaseConnection().getId() + "." + mode + "." + rt.getSchema() + "." + rt.getName() + "]");

		if (rt.getColumns().size() != 0)
		{
			return;
		}

		jdbcClient.useStatement(rt.getDatabaseConnection(), mode, new JDBCClient.StatementClient()
		{
			public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception
			{
				if (!rt.sqlDataSet())
				{
					DatabaseMetaData meta = conn.getMetaData();

					// issue a select for the table and check the meta data for column info
					String sql = "SELECT * FROM " + rt.getQualifiedName() + " WHERE 1 = 2";
					ResultSet rs = st.executeQuery(sql);

					try
					{
						ResultSetMetaData rsmd = rs.getMetaData();

						for (int col = 1; col <= rsmd.getColumnCount(); col++)
						{
							String columnName = rsmd.getColumnName(col);
							int columnType = rsmd.getColumnType(col);
							int nullable = rsmd.isNullable(col);
							boolean autoIncrement = rsmd.isAutoIncrement(col);

							RelationalColumn rc = new RelationalColumn(
									columnName,
									columnType,
									nullable == ResultSetMetaData.columnNullable,
									autoIncrement
							);

							rt.addColumn(rc);
						}
					}
					finally
					{
						rs.close();
					}

					// add all primary key columns (only the first is of interest
					rs = meta.getPrimaryKeys(null, rt.getSchema(), rt.getName());

					try
					{
						String firstPk = null;

						while (rs.next())
						{
							String columnName = rs.getString(4);
							String pkName = rs.getString(6);

							if (firstPk == null)
							{
								firstPk = pkName;
							}
							else if (!firstPk.equals(pkName))
							{
								// we don't care about the rest
								break;
							}

							rt.addPrimaryKeyColumn(rt.getColumn(columnName));
						}
					}
					finally
					{
						rs.close();
					}
				}
				else
				{
					// in the case of a SQL query, execute the query once and analyze the result set meta data
					ResultSet rs = st.executeQuery(rt.createSQLSelect());

					ResultSetMetaData rsmd = rs.getMetaData();

					try
					{
						for (int i = 1; i <= rsmd.getColumnCount(); i++)
						{
							String columnName = rsmd.getColumnName(i);
							int columnType = rsmd.getColumnType(i);

							// for the sql data set, there is no primary key
							RelationalColumn rc = new RelationalColumn(columnName, columnType, true, false);

							rt.addColumn(rc);
						}
					}
					finally
					{
						rs.close();
					}
				}
			}
		});

		// look for the meta info in the project
		try
		{
			DataFileSchema ffs = databaseRuntimeSupport.getRelationalMetaInfo(rt.getDatabaseConnection(), rt.getSchema(), rt.getName(), null, null);

			applicationLog.info("Reading meta data from reference file");
			// load the file and override the order by columns

			rt.resetOrderColumns();
			for (String name : ffs.getOrderColumnNames())
			{
				rt.addOrderByColumn(rt.getColumn(name));
			}

			rt.resetPrimaryKeyColumns();
			for (String name : ffs.getKeyColumnNames())
			{
				rt.addPrimaryKeyColumn(rt.getColumn(name));
			}

			rt.setFileSchema(ffs);
		}
		catch(ResourceNotFoundException exc)
		{
			DataFileSchema ffs = databaseRuntimeSupport.createRelationalMetaInfo(rt.getDatabaseConnection(), rt.getSchema(), rt.getName());

			applicationLog.info("Could not find reference file - creating boilerplate: " + ffs.getId());

			// generate a source definition for this set

			// write out a json object which defines the connection id and table, and the columns
			ffs.setFormatType(DataFileSchema.format_type.delimited);

			for (RelationalColumn col : rt.getColumns())
			{
				DataFileSchema.Column sc = ffs.createColumn(col.getName());
				sc.setType(JdbcTypeConverter.getTypeName(col.getType()));

				switch (col.getType())
				{
					case Types.BIGINT:
					case Types.INTEGER:
					case Types.SMALLINT:
					case Types.TINYINT:
					case Types.BIT:
						sc.setBasicType(DataFileSchema.Column.basic_type.integer);
						break;
					case Types.DECIMAL:
					case Types.DOUBLE:
					case Types.FLOAT:
					case Types.NUMERIC:
						sc.setBasicType(DataFileSchema.Column.basic_type.numeric);
						break;
				}

				ffs.addColumn(sc);
			}

			for (String colName : rt.getPrimaryKeyColumnNames())
			{
				ffs.addKeyColumn(colName);
			}

			for (String colName : rt.getOrderByColumnNames())
			{
				ffs.addKeyColumn(colName);
			}

			rt.setFileSchema(ffs);

			// load into Jackson and pretty-print
			databaseRuntimeSupport.persistDataFileSchema(ffs);
		}
	}

	private Map<String, RelationalDataSet> getTableMetaData(DatabaseConnection conn, String mode, final DatabaseImplementation impl, VariableContext context) throws Exception
	{
		String key = "database.tableSchema." + conn.getId();

		if (context.getTopLevelScope().hasVariableBeenDeclared(key))
		{
			return (Map<String, RelationalDataSet>) context.getTopLevelScope().getValue(key).getValueAsPojo();
		}

		final Map<String, RelationalDataSet> map = new HashMap<String, RelationalDataSet>();

		jdbcClient.useConnection(conn, mode, new JDBCClient.ConnectionClient()
		{
			public void connection(Connection conn, DatabaseConnection connection, String mode, int id) throws Exception
			{
				DatabaseMetaData meta = conn.getMetaData();

				ResultSet rs = meta.getTables(null, null, null, null);

				try
				{
					while (rs.next())
					{
						String tableSchema = rs.getString(2);
						String tableName = rs.getString(3);

						RelationalDataSet rt = new RelationalDataSet(tableSchema, tableName, connection, impl);
						String key1 = (connection.getId() + "." + tableSchema + '.' + tableName).toUpperCase();
						applicationLog.debug(key1);
						map.put(key1, rt);
					}
				}
				finally
				{
					rs.close();
				}
			}
		});


		context.getTopLevelScope().declareAndSetValue(key, new ETLTestValueObjectImpl(map));

		return map;
	}

	private void stage(final ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject1, StageOperation stageOperation, VariableContext context, ExecutionContext econtext) throws TestExecutionError, TestAssertionFailure, TestWarning
	{
		// check the runtime option for refresh data.  If so, convert this call into an extract operation
		if (refreshStageData)
		{
			ETLTestValueObject copy = etlTestValueObject1.copy();

			ETLTestValueObjectBuilder cbuilder = new ETLTestValueObjectBuilder(copy);

			//!!
			cbuilder = cbuilder.key("source-table").value(etlTestValueObject1.query("target-table"));
			//!!
			cbuilder.removeKey("target-table");

			if (etlTestValueObject1.query("target-schema") != null)
			{
				cbuilder = cbuilder.key("source-schema").value(etlTestValueObject1.query("target-schema"));
				cbuilder.removeKey("target-schema");
			}

			//!!
			cbuilder = cbuilder.key("target").value(etlTestValueObject1.query("source"));
			//!!
			cbuilder.removeKey("source");

			// check for multiple modes specified.  If present, replace with a single mode reference
			if (etlTestValueObject1.query("modes") != null)
			{
				cbuilder.removeKey("modes").key("mode").value(etlTestValueObject1.query("modes").getValueAsList().get(0));
			}

			copy = cbuilder.toObject();

			ETLTestOperation extract_op = etlTestOperation.createSibling("extract", copy);

			econtext.process(extract_op, context);

			return;
		}

		// look for connections and connection-id or {mode, modes}
		if (stageOperation.getConnections() != null &&
				(stageOperation.getMode() != null || stageOperation.getModes() != null || stageOperation.getConnectionId() != null)
				)
		{
			throw new TestExecutionError("Attribute " + StageOperation.CONNECTIONS_JSON_NAME + " may not be used in conjunction with  " + StageOperation.CONNECTIONS_JSON_NAME + ", " + StageOperation.MODE_JSON_NAME + " or " + StageOperation.MODES_JSON_NAME);
		}

		final String sourceFile = stageOperation.getSourceFile();
		String source = stageOperation.getSource();
		String targetTable = stageOperation.getTargetTable();

		// look for optional data
		List<ConnectionMode> connectionModes = getConnection(parent, etlTestValueObject1, context, etlTestOperation.getTestMethod());

		for (final ConnectionMode connectionMode : connectionModes)
		{
			// now send it off to the implementation
			final DatabaseImplementation impl = getDatabaseImplementation(parent, connectionMode.conn);

			File dataFile = null;

			if (sourceFile != null)
			{
				dataFile = new File(sourceFile);
			}
			else
			{
				dataFile = databaseRuntimeSupport.getSourceDataSet(etlTestOperation.getTestClass().getPackage(), source, DatabaseImplementation.data_format.delimited);
			}

			// get the relational table definition
			List<String> modeList = connectionMode.modes;

			if (modeList == null)
			{
				modeList = new ArrayList<String>();
				modeList.add(null);
			}

			for (final String thisMode : modeList)
			{
				final String schema;

				String targetSchema = stageOperation.getTargetSchema();
				if (targetSchema != null)
				{
					schema = targetSchema;
				}
				else
				{
					schema = impl.getDefaultSchema(connectionMode.conn, thisMode);
				}

				final String key = (connectionMode.connectionId + "." + schema.toUpperCase() + "." + targetTable).toUpperCase();

				try
				{
					Map<String, RelationalDataSet> meta = getTableMetaData(connectionMode.conn, thisMode, impl, context);
					final RelationalDataSet rt = meta.get(key);

					// load column and key data if not loaded already
					if (rt == null)
					{
						throw new TestExecutionError("Database does not export table " + key + " in it's meta data", DatabaseConstants.ERR_MISSING_SCHEMA_TABLE);
					}

					// make sure it is populated
					loadTable(rt, thisMode);

					applicationLog.info("Loading (" + key + ") data set (" + sourceFile + ")");

					// get the actual data file
					if (!dataFile.exists())
					{
						throw new TestExecutionError(dataFile.getAbsolutePath(), DatabaseConstants.ERR_MISSSING_DATA_FILE);
					}

					DataFileSchema fileSchema = loadSchema(rt, IOUtils.removeExtension(dataFile), etlTestValueObject1);

					final DataFile ff = dataFileManager.loadDataFile(dataFile, fileSchema);

					// open a jdbc connection and run away with it
					jdbcClient.usePreparedStatement(connectionMode.conn, thisMode, new JDBCClient.PreparedStatementClient()
					{
						public String prepareText()
						{
							String sqlInsert = rt.createSQLInsert();
							applicationLog.info("Using sql insert for object(" + key + ") in  [" + connectionMode.connectionId + "." + thisMode + "] - data set (" + sourceFile + "):\n" + sqlInsert);

							return sqlInsert;
						}

						public void connection(Connection conn, PreparedStatement pst, DatabaseConnection connection, String mode, int id) throws Exception
						{
							impl.prepareConnectionForInsert(conn, rt, connection, mode);

							List<RelationalColumn> columns = rt.getColumns();

							DataFile.FileData fdata = ff.getFileData();

							try
							{
								Iterator<DataFile.FileRow> dataIt = fdata.iterator();

								int rowNum = 0;
								int rowsInserted = 0;
								long startTime = System.currentTimeMillis();

								while (dataIt.hasNext())
								{
									rowNum++;
									rowsInserted++;

									Map<String, String> rowData = dataIt.next().getData();

									for (int i = 0; i < columns.size(); i++)
									{
										RelationalColumn col = columns.get(i);

										int columnType = col.getType();

										String val = rowData.get(col.getName());

										// TMI!!
										//applicationLog.info("Processing value: " + val);

										int parameterIndex = i + 1;
										if (val == null)
										{
											// set this to null in the insert
											pst.setNull(parameterIndex, columnType);
										}
										else
										{
											try
											{
												switch (columnType)
												{
													case Types.INTEGER:
														pst.setInt(parameterIndex, Integer.parseInt(val));
														break;
													case Types.BIGINT:
														pst.setLong(parameterIndex, Long.parseLong(val));
														break;
													case Types.SMALLINT:
														pst.setShort(parameterIndex, Short.parseShort(val));
														break;
													case Types.TINYINT:
														pst.setByte(parameterIndex, Byte.parseByte(val));
														break;
													case Types.VARCHAR:
													case Types.CHAR:
													case Types.LONGNVARCHAR:
														pst.setString(parameterIndex, val);
														break;
													case Types.CLOB:
														pst.setString(parameterIndex, val);
														break;
													case Types.TIME:
														pst.setTime(parameterIndex, Time.valueOf(val));
														break;
													case Types.DATE:
														pst.setDate(parameterIndex, Date.valueOf(val));
														break;
													case Types.TIMESTAMP:
														pst.setTimestamp(parameterIndex, Timestamp.valueOf(val));
														break;
													case Types.DECIMAL:
														pst.setBigDecimal(parameterIndex, new BigDecimal(val));
														break;
													case Types.NUMERIC:
														pst.setBigDecimal(parameterIndex, new BigDecimal(val));
														break;
													case Types.FLOAT:
													case Types.REAL:
														pst.setFloat(parameterIndex, Float.valueOf(val));
														break;
													case Types.DOUBLE:
														pst.setDouble(parameterIndex, Double.valueOf(val));
														break;
													case Types.BIT:
														pst.setBoolean(parameterIndex, Boolean.valueOf(val));
														break;
													default:
														throw new UnsupportedOperationException("Unsupported sql type: " + columnType + " java.sql.Types." + JdbcTypeConverter.getTypeName(new Integer(columnType)));
												}
											}
											catch (Throwable thr)
											{
												throw new TestExecutionError("Error loading row[" + rowNum + "], columnIndex=[" + i + "], column=[" + col + "], colType[" + columnType + "], value='" + val + "'", DatabaseConstants.ERR_DATA_TYPE_CONVERSION, thr);
											}
										}
									}

									// add as a batch
									pst.addBatch();

									if (rowNum >= 5000)
									{
										pst.executeBatch();
										rowNum = 0;
									}
								}

								if (rowNum != 0)
								{
									pst.executeBatch();
								}

								long endTime = System.currentTimeMillis();
								applicationLog.info("Inserted [" + rowsInserted + "] rows in [" + (endTime - startTime) + "] ms.");
							}
							finally
							{
								fdata.dispose();
							}
						}
					});
				}
				catch (TestExecutionError e)
				{
					throw e;
				}
				catch (Exception e)
				{
					throw new TestExecutionError("", e);
				}
			}
		}
	}

	static DatabaseImplementation getDatabaseImplementation(DatabaseFeatureModule parent, DatabaseConnection conn) throws TestExecutionError
	{
		DatabaseImplementation impl = parent.getImplementation(conn.getId());

		if (impl == null)
		{
			throw new TestExecutionError("Database implementation id " + conn.getId() + " doesn't exist", DatabaseConstants.ERR_MISSING_DATABASE_IMPLEMENTATION);
		}
		return impl;
	}

	static void validateMode(DatabaseConnection conn, String mode, VariableContext context) throws TestExecutionError
	{
		// validate that this connection was prepared for this test
		if (!context.hasVariableBeenDeclared("database"))
		{
			throw new TestExecutionError("Database support not available for this test", DatabaseConstants.ERR_MISSING_DATABASE_SUPPORT);
		}

		ETLTestValueObject database = context.getValue("database");

		ETLTestValueObject connQuery = database.query("connections." + conn.getId() + "." + (mode == null ? "default" : mode));

		if (connQuery == null)
		{
			throw new TestExecutionError("Database not prepared for connection id " + conn.getId() + ", mode " + mode, DatabaseConstants.ERR_DATABASE_MODE_NOT_PREPARED);
		}
	}

	static String getDefaultMode(DatabaseConnection conn, ETLTestMethod method) throws TestExecutionError
	{
		String mode = null;

		// validate the mode
		// the search order for modes is like this:
		// 1 - Look at method annotations.  If this connection is present, use the first mode declared
		// 2 - Look at class annotations.  If this connection is present, use the first mode declared
		List<ETLTestAnnotation> annot = new ArrayList<ETLTestAnnotation>();

		annot.addAll(method.getAnnotations("@Database"));
		annot.addAll(method.getTestClass().getAnnotations("@Database"));

		for (ETLTestAnnotation ann : annot)
		{
			ETLTestValueObject annValue = ann.getValue();
			if (annValue.query("id").getValueAsString().equals(conn.getId()))
			{
				ETLTestValueObject modes = annValue.query("modes");
				if (modes != null)
				{
					mode = modes.getValueAsList().get(0).getValueAsString();

					break;
				}
			}
		}

		return mode;
	}

	static class ConnectionMode
	{
		String connectionId;
		DatabaseConnection conn;
		List<String> modes = new ArrayList<String>();

		String getMode()
		{
			if (modes != null)
			{
				return modes.get(0);
			}

			return null;
		}

		String getPrettyMode()
		{
			String mode = getMode();

			return mode != null ? mode : "default";
		}
	}

	static List<ConnectionMode> getConnection(DatabaseFeatureModule parent, ETLTestValueObject etlTestValueObject, VariableContext context, ETLTestMethod method) throws TestExecutionError
	{
		List<ConnectionMode> connModes = new ArrayList<ConnectionMode>();

		DatabaseConfiguration databaseConfiguration = parent.getDatabaseConfiguration();

		// query for 'connections'
		ETLTestValueObject connectionsObj = etlTestValueObject.query("connections");

		if (connectionsObj != null)
		{
			Map<String, ETLTestValueObject> valueAsMap = connectionsObj.getValueAsMap();
			for (Map.Entry<String, ETLTestValueObject> connectionObj : valueAsMap.entrySet())
			{
				ConnectionMode connectionMode = new ConnectionMode();

				connectionMode.connectionId = connectionObj.getKey();

				// we either have a 'mode' here, or 'modes'
				readModes(connectionMode, connectionObj.getValue());

				connModes.add(connectionMode);
			}
		}
		else
		{
			ConnectionMode connMode = new ConnectionMode();
			connModes.add(connMode);

			// determine the connection id and mode
			ETLTestValueObject connectionIdObj = etlTestValueObject.query("connection-id");

			final DatabaseConnection conn;

			if (connectionIdObj != null)
			{
				connMode.connectionId = connectionIdObj.getValueAsString();
			}
			else
			{
				// go through the search order:
				// 1 - Annotation on the method
				// 2 - Annotation on the class
				// 3 - The default connection
				List<ETLTestAnnotation> annot = method.getAnnotations("@Database");

				if (annot.size() != 0)
				{
					// this is the one - use the first
					connMode.connectionId = annot.get(0).getValue().query("id").getValueAsString();
				}
				else
				{
					// move on to the class annotations
					annot = method.getTestClass().getAnnotations("@Database");

					if (annot.size() != 0)
					{
						// this is the one - use the first
						connMode.connectionId = annot.get(0).getValue().query("id").getValueAsString();
					}
					else
					{
						// move on to the default
						// bail
						if (!databaseConfiguration.hasDefaultConnection())
						{
							throw new TestExecutionError("Connection not specified and no default has been set", DatabaseConstants.ERR_NO_CONNECTION_SPECIFIED);
						}

						connMode.connectionId = databaseConfiguration.getDefaultConnection().getId();
					}
				}
			}

			// now check for the mode(s)
			readModes(connMode, etlTestValueObject);
		}

		// now that that's done, look up the info in the configuration for validation and set the default mode if not specified
		Iterator<ConnectionMode> it = connModes.iterator();

		while (it.hasNext())
		{
			ConnectionMode cMode = it.next();

			cMode.conn = databaseConfiguration.getDatabaseConnection(cMode.connectionId);

			if (cMode.conn == null)
			{
				throw new TestExecutionError("Connection " + cMode.conn + " is not valid", DatabaseConstants.ERR_INVALID_DATABASE_CONNECTION);
			}

			if (cMode.modes.size() == 0)
			{
				// resolve the modes for the connection
				cMode.modes.add(getDefaultMode(cMode.conn, method));
			}

			for (String mode : cMode.modes)
			{
				validateMode(cMode.conn, mode, context);
			}
		}

		return connModes;
	}

	private static void readModes(ConnectionMode connMode, ETLTestValueObject etlTestValueObject) throws TestExecutionError
	{
		// look for multiple mode and modes attributes
		if (etlTestValueObject.query(StageOperation.MODE_JSON_NAME) != null && etlTestValueObject.query(StageOperation.MODES_JSON_NAME) != null)
		{
			throw new TestExecutionError("Attributes " + StageOperation.MODE_JSON_NAME + " and " + StageOperation.MODES_JSON_NAME + " are exclusive");
		}

		ETLTestValueObject modeObj = etlTestValueObject.query("mode");

		if (modeObj != null)
		{
			connMode.modes.add(modeObj.getValueAsString());
		}
		else
		{
			ETLTestValueObject modesObj = etlTestValueObject.query("modes");

			if (modesObj != null)
			{
				for (String str : modesObj.getValueAsListOfStrings())
				{
					connMode.modes.add(str);
				}
			}
		}
	}

	public Feature getFeature()
	{
		return parent;
	}

	public action_code processStage(StageOperation operation, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		stage(op, obj, operation, context, econtext);

		return action_code.handled;
	}

	public action_code processTruncate(TruncateOperation operation, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		final String target = operation.getTarget();

		final ConnectionMode connMode = getConnection(parent, obj, context, op.getTestMethod()).get(0);

		DatabaseConnection conn = connMode.conn;

		String mode = connMode.getMode();

		DatabaseImplementation impl = getDatabaseImplementation(parent, conn);

		final String schema = operation.getSchema() == null ? impl.getDefaultSchema(conn, mode) : operation.getSchema();

		// get the connection and blast the sucker
		jdbcClient.useStatement(conn, mode, new JDBCClient.StatementClient()
		{
			public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception
			{
				String sql = "DELETE FROM " + schema + "." + target;
				applicationLog.info("Using truncate sql [" + connection.getId() + "." + connMode.getPrettyMode() + "]:\n" + sql);
				st.executeUpdate(sql);
			}
		});

		return action_code.handled;
	}

	class MigrateInfo
	{
		String sourceValue;
		String targetValue;
	}

	class MigrateInfos
	{
		MigrateInfo tableInfo;
		MigrateInfo connectionInfo;
		MigrateInfo schemaInfo;
		MigrateInfo modeInfo;
	}

	public action_code processMigrate(MigrateOperation operation, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		MigrateInfos mi = readMigrateExclusiveOptions(operation);

		// this process is three steps:
		// 1 - run truncate on the target
		// 2 - run extract on the source
		// 3 - run stage on the target
		ETLTestValueObject truncateValueObj = obj.copy();

		/* truncate has to see:
		-- target    << target-table
		-- optionally:
		--  connection-id << passes through
		--  schema
		--  mode
		-- for this purpose, we will need to convert the target-* properties to
		-- the unqualified properties truncate wants to see.
		*/

		ETLTestValueObjectBuilder builder = new ETLTestValueObjectBuilder(truncateValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("target").value(mi.tableInfo.targetValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.targetValue != null)
		{
			builder = builder.key("connection-id").value(mi.connectionInfo.targetValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.targetValue != null)
		{
			builder = builder.key("schema").value(mi.schemaInfo.targetValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.targetValue != null)
		{
			builder = builder.key("mode").value(mi.modeInfo.targetValue);
		}

		// create the operation and dispatch
		ETLTestOperation truncateOp = op.createSibling("truncate", truncateValueObj);

		econtext.process(truncateOp, context);

		/* extract needs:
		-- source-table *
		-- connection-id *
		-- mode *
		-- target-file *
		-- source-schema *
		-- column-list *
		-- column-list-mode *
		 */

		ETLTestValueObject extractValueObj = obj.copy();
		builder = new ETLTestValueObjectBuilder(extractValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("source-table").value(mi.tableInfo.sourceValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.sourceValue != null)
		{
			builder = builder.key("connection-id").value(mi.connectionInfo.sourceValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.sourceValue != null)
		{
			builder = builder.key("source-schema").value(mi.schemaInfo.sourceValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.sourceValue != null)
		{
			builder = builder.key("mode").value(mi.modeInfo.sourceValue);
		}

		ConnectionMode connMode = getConnection(parent, extractValueObj, context, op.getTestMethod()).get(0);

		DatabaseConnection conn = connMode.conn;

		File targetFile = databaseRuntimeSupport.getGeneratedDataSet(conn, connMode.getPrettyMode(), op.getQualifiedName() + "_extract", DatabaseImplementation.data_format.delimited);

		// remove all mode attributes and set
		builder = builder.key("target-file").value(targetFile.getAbsolutePath());

		// create the operation and dispatch
		ETLTestOperation extractOp = op.createSibling("extract", extractValueObj);

		econtext.process(extractOp, context);

		/* stage needs:
		-- source-file *
		-- target-table *
		-- connection-id *
		-- target-schema *
		-- mode *
		 */
		ETLTestValueObject stageValueObj = obj.copy();
		builder = new ETLTestValueObjectBuilder(stageValueObj);

		// remove all table attributes and set
		builder = builder.removeAllKeys("source-table", "target-table", "table");
		builder = builder.key("source-file").value(targetFile.getAbsolutePath());
		builder = builder.key("target-table").value(mi.tableInfo.targetValue);

		// remove all connection attributes and set
		builder = builder.removeAllKeys("source-connection", "target-connection", "connection");
		if (mi.connectionInfo.targetValue != null)
		{
			builder = builder.key("connection-id").value(mi.connectionInfo.targetValue);
		}

		// remove all schema attributes and set
		builder = builder.removeAllKeys("source-schema", "target-schema", "schema");
		if (mi.schemaInfo.targetValue != null)
		{
			builder = builder.key("source-schema").value(mi.schemaInfo.targetValue);
		}

		// remove all mode attributes and set
		builder = builder.removeAllKeys("source-mode", "target-mode", "mode");
		if (mi.modeInfo.targetValue != null)
		{
			builder = builder.key("mode").value(mi.modeInfo.targetValue);
		}

		// create the operation and dispatch
		ETLTestOperation stageOp = op.createSibling("stage", stageValueObj);

		econtext.process(stageOp, context);

		return action_code.handled;
	}

	private MigrateInfos readMigrateExclusiveOptions(MigrateOperation operation) throws TestExecutionError
	{
		// table
		String table = operation.getTable();
		String sourceTable = operation.getSourceTable();
		String targetTable = operation.getTargetTable();

		// connection ID
		String conId = operation.getConnectionId();
		String sourceConId = operation.getSourceConnectionId();
		String targetConId = operation.getTargetConnectionId();

		// schema
		String schema = operation.getSchema();
		String sourceSchema = operation.getSourceSchema();
		String targetSchema = operation.getTargetSchema();

		// mode
		String mode = operation.getMode();
		String sourceMode = operation.getSourceMode();
		String targetMode = operation.getTargetMode();

		MigrateInfos mi = new MigrateInfos();

		mi.tableInfo = checkExclusive(table, sourceTable, targetTable, "table");
		mi.connectionInfo = checkExclusive(conId, sourceConId, targetConId, "connection-id");
		mi.schemaInfo = checkExclusive(schema, sourceSchema, targetSchema, "schema");
		mi.modeInfo = checkExclusive(mode, sourceMode, targetMode, "mode");

		/* there are a few error conditions here:
		-- 1 - If the source and target tables are the same, meaning
		       that the table, connection, schema and mode are all the same
		*/

		if (
				compare(mi.tableInfo.sourceValue, mi.tableInfo.targetValue)
						&&
						compare(mi.connectionInfo.sourceValue, mi.connectionInfo.targetValue)
						&&
						compare(mi.schemaInfo.sourceValue, mi.schemaInfo.targetValue)
						&&
						compare(mi.modeInfo.sourceValue, mi.modeInfo.targetValue))
		{
			throw new TestExecutionError("Source and target tables are the same", DatabaseConstants.ERR_SOURCE_TARGET_IDENTICAL);
		}

		return mi;
	}

	private boolean compare(String s, String t)
	{
		if (s == null && t == null)
		{
			return true;
		}
		else if (s == null && t != null)
		{
			return false;
		}
		else if (s != null && t == null)
		{
			return false;
		}
		else if (s == t)
		{
			return true;
		}
		else if (s.equals(t))
		{
			return true;
		}

		return false;
	}

	private MigrateInfo checkExclusive(String conId, String sourceConId, String targetConId, String baseAttribute) throws TestExecutionError
	{
		MigrateInfo mt = new MigrateInfo();

		if (conId != null || sourceConId != null || targetConId != null)
		{
			// if the generic is specified and either of the specific ones is specified, fail
			if (conId != null && (sourceConId != null || targetConId != null))
			{
				throw new TestExecutionError("When " + baseAttribute + " is specified, neither of source-" + baseAttribute + " or target-" + baseAttribute + " may be specified", DatabaseConstants.ERR_GENERIC_AND_SPECIFIC_OPTIONS_FAIL);
			}

			// if the generic is NOT specified, then both of the specific operands are required
			if (conId == null && (sourceConId == null || targetConId == null))
			{
				throw new TestExecutionError("When " + baseAttribute + " is not specified, both source-" + baseAttribute + " and target-" + baseAttribute + " must be specified", DatabaseConstants.ERR_GENERIC_AND_SPECIFIC_OPTIONS_FAIL);
			}

			mt.sourceValue = conId == null ? sourceConId : conId;
			mt.targetValue = conId == null ? targetConId : conId;
		}

		return mt;
	}
}